searchNodes=[{"doc":"RTC Engine implementation. RTC Engine is an abstraction layer responsible for linking together different types of Endpoints . From the implementation point of view, RTC Engine is a Membrane.Pipeline . Messages The RTC Engine works by sending messages which notify user logic about important events like &quot;There is a new peer, would you like to to accept it?&quot;. To receive RTC Engine messages you have to register your process so that RTC Engine will know where to send them. All messages RTC Engine can emit are described in Membrane.RTC.Engine.Message docs. Registering for messages Registration can be done using register/2 e.g. Engine . register ( rtc_engine , self ( ) ) This will register your process to receive RTC Engine messages. If your process implements GenServer behavior then all messages can be handled by GenServer.handle_info/2 , e.g. @impl true def handle_info ( % Message.NewPeer { rtc_engine : rtc_engine , peer : peer } , state ) do Engine . accept_peer ( rtc_engine , peer . id ) { :noreply , state } end You can register multiple processes to receive messages from an RTC Engine instance. In such a case each message will be sent to each registered process. Client Libraries RTC Engine allows creating Client Libraries that can send and receive media tracks from it. The current version of RTC Engine ships with WebRTC Client Library which connects to the RTC Engine via WebRTC standard. Communication with Client Libraries is done using Media Events . Media Events are control messages which notify about e.g. new peer joining to the RTC Engine. When Client Library receives Media Event it can invoke some callbacks. In the case of WebRTC Client Library, these are e.g. onPeerJoined or onTrackAdded . When RTC Engine receives Media Event it can emit some messages e.g. Membrane.RTC.Engine.Message.NewPeer.t/0 . More about Media Events can be read in subsequent sections. Below there is a figure showing the architecture of the RTC Engine working in conjunction with some Client Library. +--------------------------------- media events -----------------------------+ | (signaling layer) | | | | | +--------+ +---------+ +--------+ +---------+ | user | &lt;- media -&gt; | Client | | RTC | &lt;- media -&gt; | user | | client | events | Library | &lt;- media -&gt; | Engine | events | backend | | logic | &lt;- callbacks - | | | | - messages -&gt; | logic | +--------+ +---------+ +--------+ +---------+ Media Events Media Events are blackbox messages that carry data important for the RTC Engine and its Client Library, but not for the user. There are two types of Media Events: Internal Media Events - generic, protocol-agnostic Media Events sent by RTC Engine itself. Example Internal Media Events are peerJoined , peerLeft , tracksAdded or tracksRemoved . Custom Media Events - they can be used to send custom data from Client Library to some Endpoint inside RTC Engine and vice versa. In the case of WebRTC Client Library, these are sdpOffer , sdpAnswer , or iceCandidate . An application is obligated to transport Media Events from an RTC Engine instance to its Client Library, and vice versa. When the RTC Engine needs to send a Media Event to a specific client, registered processes will receive Membrane.RTC.Engine.Message.MediaEvent.t/0 message with to field indicating where this Media Event should be sent to. This can be either :broadcast , when the event should be sent to all peers, or peer_id when the messages should be sent to the specified peer. The event is encoded in binary format, so it is ready to send without modification. Feeding an RTC Engine instance with Media Events from a Client Library can be done using receive_media_event/2 . Assuming the user process is a GenServer, the Media Event can be received by GenServer.handle_info/2 and conveyed to the RTC Engine in the following way: @impl true def handle_info ( { :media_event , from , event } = msg , state ) do Engine . receive_media_event ( state . rtc_engine , from , event ) { :noreply , state } end What is important, Membrane RTC Engine doesn't impose usage of any specific transport layer for carrying Media Events through the network. You can e.g. use Phoenix and its channels. This can look like this: @impl true def handle_in ( &quot;mediaEvent&quot; , %{ &quot;data&quot; =&gt; event } , socket ) do Engine . receive_media_event ( socket . assigns . room , socket . assigns . peer_id , event ) { :noreply , socket } end Peers Each peer represents some user that can possess some metadata. A Peer can be added in two ways: by sending proper Media Event from a Client Library using add_peer/3 Adding a peer will cause RTC Engine to emit Media Event which will notify connected clients about new peer. Peer id Peer ids must be assigned by application code. This is not done by the RTC Engine or its client library. Ids can be assigned when a peer initializes its signaling layer. Endpoints Endpoints are Membrane.Bin s able to publish their own tracks and subscribe for tracks from other Endpoints. One can think about Endpoint as an entity responsible for handling some specific task. An Endpoint can be added and removed using add_endpoint/3 and remove_endpoint/2 respectively. There are two types of Endpoints: Standalone Endpoints - they are in most cases spawned only once per RTC Engine instance and they are not associated with any peer. Peer Endpoints - they are associated with some peer. Associating Endpoint with Peer will cause RTC Engine to send some Media Events to the Enpoint's Client Library e.g. one which indicates which tracks belong to which peer. Currently RTC Engine ships with the implementation of two Endpoints: Membrane.RTC.Engine.Endpoint.WebRTC which is responsible for establishing a connection with some WebRTC peer (mainly browser) and exchanging media with it. WebRTC Endpoint is a Peer Endpoint. Membrane.RTC.Engine.Endpoint.HLS which is responsible for receiving media tracks from all other Endpoints and saving them to files by creating HLS playlists. HLS Endpoint is a Standalone Endpoint. User can also implement custom Endpoints. Implementing custom RTC Engine Endpoint Each RTC Engine Endpoint has to: implement Membrane.Bin behavior specify input, output, or both input and output pads depending on what it is intended to do. For example, if Endpoint will not publish any tracks but only subscribe for tracks from other Endpoints it can specify only input pads. Pads should have the following form def_input_pad :input , demand_unit : :buffers , caps : &lt; caps &gt; , availability : :on_request def_output_pad :output , demand_unit : :buffers , caps : &lt; caps &gt; , availability : :on_request Where caps are Membrane.Caps.t/0 or :any . publish for some tracks using actions publish_action_t/0 and subscribe for some tracks using function Membrane.RTC.Engine.subscribe/5 . The first will cause RTC Engine to send a message in form of {:new_tracks, tracks} where tracks is a list of Membrane.RTC.Engine.Track.t/0 to all other Endpoints. When an Endpoint receives such a message it can subscribe for new tracks by using Membrane.RTC.Engine.subscribe/5 function. An Endpoint will be notified about track readiness it subscribed for in Membrane.Bin.handle_pad_added/3 callback. An example implementation of handle_pad_added callback can look like this @impl true def handle_pad_added ( Pad . ref ( :input , _track_id ) = pad , _ctx , state ) do links = [ link_bin_input ( pad ) |&gt; via_in ( pad ) |&gt; to ( :my_element ) ] { { :ok , spec : % ParentSpec { links : links } } , state } end Where :my_element is a custom Membrane element responsible for processing track. Endpoint will be also notified when some tracks it subscribed for are removed with {:removed_tracks, tracks} message where tracks is a list of Membrane.RTC.Engine.Track.t/0 .","ref":"Membrane.RTC.Engine.html","title":"Membrane.RTC.Engine","type":"module"},{"doc":"Allows peer for joining to the RTC Engine","ref":"Membrane.RTC.Engine.html#accept_peer/2","title":"Membrane.RTC.Engine.accept_peer/2","type":"function"},{"doc":"Adds endpoint to the RTC Engine Returns :error when there are both peer_id and endpoint_id specified in opts . For more information refer to endpoint_options_t/0 .","ref":"Membrane.RTC.Engine.html#add_endpoint/3","title":"Membrane.RTC.Engine.add_endpoint/3","type":"function"},{"doc":"Adds peer to the RTC Engine","ref":"Membrane.RTC.Engine.html#add_peer/2","title":"Membrane.RTC.Engine.add_peer/2","type":"function"},{"doc":"Deny peer from joining to the RTC Engine.","ref":"Membrane.RTC.Engine.html#deny_peer/2","title":"Membrane.RTC.Engine.deny_peer/2","type":"function"},{"doc":"The same as deny_peer/2 but allows for passing any data that will be returned to the client. This can be used for passing reason of peer refusal.","ref":"Membrane.RTC.Engine.html#deny_peer/3","title":"Membrane.RTC.Engine.deny_peer/3","type":"function"},{"doc":"","ref":"Membrane.RTC.Engine.html#get_registry_name/0","title":"Membrane.RTC.Engine.get_registry_name/0","type":"function"},{"doc":"Changes playback state of pipeline to :playing .","ref":"Membrane.RTC.Engine.html#play/1","title":"Membrane.RTC.Engine.play/1","type":"function"},{"doc":"Changes playback state to :prepared .","ref":"Membrane.RTC.Engine.html#prepare/1","title":"Membrane.RTC.Engine.prepare/1","type":"function"},{"doc":"Sends Media Event to RTC Engine.","ref":"Membrane.RTC.Engine.html#receive_media_event/2","title":"Membrane.RTC.Engine.receive_media_event/2","type":"function"},{"doc":"Registers process with pid who for receiving messages from RTC Engine","ref":"Membrane.RTC.Engine.html#register/2","title":"Membrane.RTC.Engine.register/2","type":"function"},{"doc":"Removes endpoint from the RTC Engine","ref":"Membrane.RTC.Engine.html#remove_endpoint/2","title":"Membrane.RTC.Engine.remove_endpoint/2","type":"function"},{"doc":"Removes peer from RTC Engine. If reason is other than nil , RTC Engine will inform client library about peer removal with passed reason.","ref":"Membrane.RTC.Engine.html#remove_peer/3","title":"Membrane.RTC.Engine.remove_peer/3","type":"function"},{"doc":"","ref":"Membrane.RTC.Engine.html#start/2","title":"Membrane.RTC.Engine.start/2","type":"function"},{"doc":"","ref":"Membrane.RTC.Engine.html#start_link/2","title":"Membrane.RTC.Engine.start_link/2","type":"function"},{"doc":"Changes playback state to :stopped .","ref":"Membrane.RTC.Engine.html#stop/1","title":"Membrane.RTC.Engine.stop/1","type":"function"},{"doc":"Changes pipeline's playback state to :stopped and terminates its process.","ref":"Membrane.RTC.Engine.html#stop_and_terminate/2","title":"Membrane.RTC.Engine.stop_and_terminate/2","type":"function"},{"doc":"Subscribes endpoint for tracks. Endpoint will be notified about track readiness in Membrane.Bin.handle_pad_added/3 callback. tracks is a list in form of pairs {track_id, track_format} , where track_id is id of track this endpoint subscribes for and track_format is the format of track that this endpoint is willing to receive. If track_format is :raw Endpoint will receive track in Membrane.RTC.Engine.Track.encoding/0 format. Endpoint_id is a an id of endpoint, which want to subscribe on tracks.","ref":"Membrane.RTC.Engine.html#subscribe/5","title":"Membrane.RTC.Engine.subscribe/5","type":"function"},{"doc":"Changes pipeline's playback state to :stopped and terminates its process.","ref":"Membrane.RTC.Engine.html#terminate/2","title":"Membrane.RTC.Engine.terminate/2","type":"function"},{"doc":"Unregisters process with pid who from receiving messages from RTC Engine","ref":"Membrane.RTC.Engine.html#unregister/2","title":"Membrane.RTC.Engine.unregister/2","type":"function"},{"doc":"Membrane action that will generate Custom Media Event.","ref":"Membrane.RTC.Engine.html#t:custom_media_event_action_t/0","title":"Membrane.RTC.Engine.custom_media_event_action_t/0","type":"type"},{"doc":"Endpoint configuration options. peer_id - associate endpoint with exisiting peer endpoint_id - assign endpoint id. If not provided it will be generated by RTC Engine. This option cannot be used together with peer_id . Endpoints associated with peers have the id peer_id . node - node on which endpoint should be spawned. If not provided, current node is used.","ref":"Membrane.RTC.Engine.html#t:endpoint_options_t/0","title":"Membrane.RTC.Engine.endpoint_options_t/0","type":"type"},{"doc":"RTC Engine configuration options. id is used by logger. If not provided it will be generated. trace_ctx is used by OpenTelemetry. All traces from this engine will be attached to this context. Example function from which you can get Otel Context is get_current/0 from OpenTelemetry.Ctx . display_manager? - set to true if you want to limit number of tracks sent from Membrane.RTC.Engine.Endpoint.WebRTC to a browser.","ref":"Membrane.RTC.Engine.html#t:options_t/0","title":"Membrane.RTC.Engine.options_t/0","type":"type"},{"doc":"Membrane action that will cause RTC Engine to publish some message to all other endpoints.","ref":"Membrane.RTC.Engine.html#t:publish_action_t/0","title":"Membrane.RTC.Engine.publish_action_t/0","type":"type"},{"doc":"Types of messages that can be published to other Endpoints.","ref":"Membrane.RTC.Engine.html#t:publish_message_t/0","title":"Membrane.RTC.Engine.publish_message_t/0","type":"type"},{"doc":"Subscription options. default_simulcast_encoding - initial encoding that endpoint making subscription wants to receive. This option has no effect for audio tracks and video tracks that are not simulcast.","ref":"Membrane.RTC.Engine.html#t:subscription_opts_t/0","title":"Membrane.RTC.Engine.subscription_opts_t/0","type":"type"},{"doc":"Membrane action that will inform RTC Engine about track readiness.","ref":"Membrane.RTC.Engine.html#t:track_ready_action_t/0","title":"Membrane.RTC.Engine.track_ready_action_t/0","type":"type"},{"doc":"Module describing Peer.","ref":"Membrane.RTC.Engine.Peer.html","title":"Membrane.RTC.Engine.Peer","type":"module"},{"doc":"","ref":"Membrane.RTC.Engine.Peer.html#new/2","title":"Membrane.RTC.Engine.Peer.new/2","type":"function"},{"doc":"","ref":"Membrane.RTC.Engine.Peer.html#t:id/0","title":"Membrane.RTC.Engine.Peer.id/0","type":"type"},{"doc":"","ref":"Membrane.RTC.Engine.Peer.html#t:t/0","title":"Membrane.RTC.Engine.Peer.t/0","type":"type"},{"doc":"Module representing media track. Media track is a single audio or video. Tracks that are related to each other (e.g. audio from microphone that corresponds to video from a web cam) can be grouped into the same stream by assigning each of them the same stream id.","ref":"Membrane.RTC.Engine.Track.html","title":"Membrane.RTC.Engine.Track","type":"module"},{"doc":"Creates a new track. Tracks belonging to the same stream should have the same stream_id , that can be generated with stream_id/0 .","ref":"Membrane.RTC.Engine.Track.html#new/8","title":"Membrane.RTC.Engine.Track.new/8","type":"function"},{"doc":"Generates stream id, that can be used to mark tracks belonging to the same stream.","ref":"Membrane.RTC.Engine.Track.html#stream_id/0","title":"Membrane.RTC.Engine.Track.stream_id/0","type":"function"},{"doc":"","ref":"Membrane.RTC.Engine.Track.html#t:encoding/0","title":"Membrane.RTC.Engine.Track.encoding/0","type":"type"},{"doc":"","ref":"Membrane.RTC.Engine.Track.html#t:format/0","title":"Membrane.RTC.Engine.Track.format/0","type":"type"},{"doc":"","ref":"Membrane.RTC.Engine.Track.html#t:id/0","title":"Membrane.RTC.Engine.Track.id/0","type":"type"},{"doc":"Options that can be passed to new/7 . If not provided: id - will be generated simulcast_encodings - [] metadata - nil ctx - %{} For more information refer to t/0 .","ref":"Membrane.RTC.Engine.Track.html#t:opts_t/0","title":"Membrane.RTC.Engine.Track.opts_t/0","type":"type"},{"doc":"This module contains: type - audio or video, stream_id - media stream this track belongs to. Relationship between tracks (e.g. audio and video) can be indicated by assigning each of them the same stream_id . One stream_id can be assign to any number of tracks. id - track id origin - id of Endpoint this track belongs to encoding - track encoding simulcast_encodings - for simulcast tracks this is a list of simulcast encoding identifiers. In other case, this is an empty list. clock_rate - track clock rate format - list of available track formats. At this moment max two formats can be specified. One of them has to be :raw which indicates that other Endpoints will receive this track in format of encoding . The other one can be any atom (e.g. :RTP ). fmtp - struct describing format specific parameters e.g. for H264 it contains profile_level_id active? - indicates whether track is still available or not (because peer left a room) metadata - any data passed by user to be linked with this track ctx - any data Endpoints need to associate with Membrane.RTC.Engine.Track.t() for internal usage","ref":"Membrane.RTC.Engine.Track.html#t:t/0","title":"Membrane.RTC.Engine.Track.t/0","type":"type"},{"doc":"An Endpoint responsible for converting incoming tracks to HLS playlist. This module requires the following plugins to be present in your mix.exs : membrane_http_adaptive_stream_plugin, membrane_mp4_plugin, membrane_aac_plugin, membrane_aac_fdk_plugin, Bin options Passed via struct Membrane.RTC.Engine.Endpoint.HLS.t/0 rtc_engine pid ( ) Required Pid of parent Engine output_directory Path . t ( ) Default value: &quot;hls_output&quot; Path to directory under which HLS output will be saved owner pid ( ) Required Pid of parent all notifications will be send to. These notifications are: {:playlist_playable, content_type, stream_id} {:cleanup, clean_function, stream_id} framerate { integer ( ) , integer ( ) } Default value: {30, 1} Framerate of tracks hls_mode :separate_av | :muxed_av Default value: :separate_av Defines output mode for Membrane.HTTPAdaptiveStream.SinkBin . :separate_av - audio and video tracks will be separated :muxed_av - audio will be attached to every video track Pads :input Availability :on_request Caps :any Demand unit :buffers Direction :input Mode :pull Name :input","ref":"Membrane.RTC.Engine.Endpoint.HLS.html","title":"Membrane.RTC.Engine.Endpoint.HLS","type":"module"},{"doc":"Returns pads descriptions for Membrane.RTC.Engine.Endpoint.HLS","ref":"Membrane.RTC.Engine.Endpoint.HLS.html#membrane_pads/0","title":"Membrane.RTC.Engine.Endpoint.HLS.membrane_pads/0","type":"function"},{"doc":"Returns description of options available for this module","ref":"Membrane.RTC.Engine.Endpoint.HLS.html#options/0","title":"Membrane.RTC.Engine.Endpoint.HLS.options/0","type":"function"},{"doc":"Struct containing options for Membrane.RTC.Engine.Endpoint.HLS","ref":"Membrane.RTC.Engine.Endpoint.HLS.html#t:t/0","title":"Membrane.RTC.Engine.Endpoint.HLS.t/0","type":"type"},{"doc":"An Endpoint responsible for communicatiing with WebRTC peer. It is responsible for sending and receiving media tracks from other WebRTC peer (e.g. web browser). Bin options Passed via struct Membrane.RTC.Engine.Endpoint.WebRTC.t/0 rtc_engine pid ( ) Required Pid of parent Engine ice_name String . t ( ) Required Ice name is used in creating credentials for ice connnection handshake_opts Keyword . t ( ) Default value: [] Keyword list with options for handshake module. For more information please refer to Membrane.ICE.Bin filter_codecs ( { RTPMapping . t ( ) , FMTP . t ( ) | nil } -&gt; boolean ( ) ) Default value: &amp;Membrane.WebRTC.SDP.filter_mappings/1 Defines function which will filter SDP m-line by codecs log_metadata :list Default value: [] Logger metadata used for endpoint bin and all its descendants webrtc_extensions [ Membrane.WebRTC.Extension . t ( ) ] Default value: [] List of WebRTC extensions to use. At this moment only VAD (RFC 6464) is supported. Enabling it will cause RTC Engine sending {:vad_notification, val, endpoint_id} messages. extensions %{ ( encoding_name :: atom ( ) | :any ) =&gt; [ Membrane.RTP.SessionBin . extension_t ( ) ] } Default value: %{} A map pointing from encoding names to lists of extensions that should be used for given encodings. Encoding &quot; :any &quot; indicates that extensions should be applied regardless of encoding. A sample usage would be to add silence discarder to OPUS tracks when VAD extension is enabled. It can greatly reduce CPU usage in rooms when there are a lot of people but only a few of them are actively speaking. integrated_turn_domain binary ( ) | nil Default value: nil Domain address of integrated TURN Servers. Required for TLS TURN integrated_turn_options Membrane.TURN.Endpoint . integrated_turn_options_t ( ) Default value: [] owner pid ( ) Required Pid of parent all notifications will be send to. To see possible notifications please refer to module docs. trace_context :list Default value: [] Trace context for otel propagation video_tracks_limit integer ( ) | nil Default value: nil Maximal number of video tracks that will be sent to the the browser at the same time. This variable indicates how many video tracks should be sent to the browser at the same time. If nil all video tracks this Membrane.RTC.Engine.Endpoint.WebRTC receives will be sent. rtcp_receiver_report_interval Membrane.Time . t ( ) | nil Default value: nil Receiver reports's generation interval, set to nil to avoid reports generation rtcp_sender_report_interval Membrane.Time . t ( ) | nil Default value: nil Sender reports's generation interval, set to nil to avoid reports generation rtcp_fir_interval Membrane.Time . t ( ) | nil Default value: 1000000000 Defines how often FIR should be sent. For more information refer to RFC 5104 section 4.3.1. simulcast_config SimulcastConfig . t ( ) Default value: %Membrane.RTC.Engine.Endpoint.WebRTC.SimulcastConfig{default_encoding: &amp;Membrane.RTC.Engine.Endpoint.WebRTC.SimulcastConfig.default_encoding/1, enabled: false} Simulcast configuration Pads :input Availability :on_request Caps :any Demand unit :buffers Direction :input Mode :pull Name :input :output Availability :on_request Caps :any Demand unit :buffers Direction :output Mode :pull Name :output","ref":"Membrane.RTC.Engine.Endpoint.WebRTC.html","title":"Membrane.RTC.Engine.Endpoint.WebRTC","type":"module"},{"doc":"Returns pads descriptions for Membrane.RTC.Engine.Endpoint.WebRTC","ref":"Membrane.RTC.Engine.Endpoint.WebRTC.html#membrane_pads/0","title":"Membrane.RTC.Engine.Endpoint.WebRTC.membrane_pads/0","type":"function"},{"doc":"Returns description of options available for this module","ref":"Membrane.RTC.Engine.Endpoint.WebRTC.html#options/0","title":"Membrane.RTC.Engine.Endpoint.WebRTC.options/0","type":"function"},{"doc":"","ref":"Membrane.RTC.Engine.Endpoint.WebRTC.html#t:encoding_t/0","title":"Membrane.RTC.Engine.Endpoint.WebRTC.encoding_t/0","type":"type"},{"doc":"","ref":"Membrane.RTC.Engine.Endpoint.WebRTC.html#t:stun_server_t/0","title":"Membrane.RTC.Engine.Endpoint.WebRTC.stun_server_t/0","type":"type"},{"doc":"Struct containing options for Membrane.RTC.Engine.Endpoint.WebRTC","ref":"Membrane.RTC.Engine.Endpoint.WebRTC.html#t:t/0","title":"Membrane.RTC.Engine.Endpoint.WebRTC.t/0","type":"type"},{"doc":"","ref":"Membrane.RTC.Engine.Endpoint.WebRTC.html#t:turn_server_t/0","title":"Membrane.RTC.Engine.Endpoint.WebRTC.turn_server_t/0","type":"type"},{"doc":"Module representing simulcast configuration for WebRTC endpoint.","ref":"Membrane.RTC.Engine.Endpoint.WebRTC.SimulcastConfig.html","title":"Membrane.RTC.Engine.Endpoint.WebRTC.SimulcastConfig","type":"module"},{"doc":"Default implementation of default_encoding function in t/0 . Returns nil, which will result in choosing the highest possible encoding.","ref":"Membrane.RTC.Engine.Endpoint.WebRTC.SimulcastConfig.html#default_encoding/1","title":"Membrane.RTC.Engine.Endpoint.WebRTC.SimulcastConfig.default_encoding/1","type":"function"},{"doc":"enabled - whether to accept simulcast tracks or not. Setting this to false will result in rejecting all incoming simulcast tracks i.e. client will not send them. default_encoding - function used to determine initial encoding this endpoint is willing to receive for given track. It is called for each track this endpoint subscribes for. If not provided, the highest possible encoding will be used.","ref":"Membrane.RTC.Engine.Endpoint.WebRTC.SimulcastConfig.html#t:t/0","title":"Membrane.RTC.Engine.Endpoint.WebRTC.SimulcastConfig.t/0","type":"type"},{"doc":"Module describing messages RTC Engine can emit. Each Message contains RTC Engine PID under rtc_engine field. Thanks to it you can distinguish between the same messages but from different RTC Engine instances.","ref":"Membrane.RTC.Engine.Message.html","title":"Membrane.RTC.Engine.Message","type":"module"},{"doc":"","ref":"Membrane.RTC.Engine.Message.html#t:t/0","title":"Membrane.RTC.Engine.Message.t/0","type":"type"},{"doc":"Message emitted when an endpoint crashes.","ref":"Membrane.RTC.Engine.Message.EndpointCrashed.html","title":"Membrane.RTC.Engine.Message.EndpointCrashed","type":"module"},{"doc":"Describes EndpointCrashed Message structure. endpoint_id - id of an endpoint that crashed","ref":"Membrane.RTC.Engine.Message.EndpointCrashed.html#t:t/0","title":"Membrane.RTC.Engine.Message.EndpointCrashed.t/0","type":"type"},{"doc":"Message emitted when RTC Engine need to send some Media Event to the Client Library.","ref":"Membrane.RTC.Engine.Message.MediaEvent.html","title":"Membrane.RTC.Engine.Message.MediaEvent","type":"module"},{"doc":"Describes MediaEvent Message structure. rtc_engine - pid of RTC Engine instance which emitted this message to - informs where this Media Event should be sent. If set to :broadcast , the Media Event should be sent to all peers. When set to t:Membrane.RTC.Engine.Peer.id() , the Media Event should be sent to that specified peer. data - Media Event in serialized i.e. binary form","ref":"Membrane.RTC.Engine.Message.MediaEvent.html#t:t/0","title":"Membrane.RTC.Engine.Message.MediaEvent.t/0","type":"type"},{"doc":"Message emitted when a new peer from Client Library tries to join RTC Engine. You can reply to this message using: Membrane.RTC.Engine.accept_peer/2 and Membrane.RTC.Engine.deny_peer/2 or Membrane.RTC.Engine.deny_peer/3 .","ref":"Membrane.RTC.Engine.Message.NewPeer.html","title":"Membrane.RTC.Engine.Message.NewPeer","type":"module"},{"doc":"Describes NewPeer Message structure. rtc_engine - pid of RTC Engine instance which emitted this message peer - peer that tries to join to RTC Engine","ref":"Membrane.RTC.Engine.Message.NewPeer.html#t:t/0","title":"Membrane.RTC.Engine.Message.NewPeer.t/0","type":"type"},{"doc":"Message emitted when a peer left RTC Engine.","ref":"Membrane.RTC.Engine.Message.PeerLeft.html","title":"Membrane.RTC.Engine.Message.PeerLeft","type":"module"},{"doc":"Describes PeerLeft Message structure. rtc_engine - pid of RTC Engine instance which emitted this message peer - peer that left RTC Engine","ref":"Membrane.RTC.Engine.Message.PeerLeft.html#t:t/0","title":"Membrane.RTC.Engine.Message.PeerLeft.t/0","type":"type"},{"doc":"Customizable Real-time Communication Engine/SFU library focused on WebRTC.","ref":"readme.html","title":"Membrane RTC Engine","type":"extras"},{"doc":"JS/TS Android iOS React Native","ref":"readme.html#client-sdks","title":"Membrane RTC Engine - Client SDKs","type":"extras"},{"doc":"The package can be installed by adding membrane_rtc_engine to your list of dependencies in mix.exs : def deps do [ { :membrane_rtc_engine , &quot;~&gt; 0.3.0&quot; } ] end","ref":"readme.html#installation","title":"Membrane RTC Engine - Installation","type":"extras"},{"doc":"For usage examples, please refer to our membrane_demo or membrane_videoroom repositories.","ref":"readme.html#usage","title":"Membrane RTC Engine - Usage","type":"extras"},{"doc":"To make the development a little easier, we have added mix integration_test task, which runs integration tests from integration/test_videoroom .","ref":"readme.html#developing","title":"Membrane RTC Engine - Developing","type":"extras"},{"doc":"Copyright 2021, Software Mansion Licensed under the Apache License, Version 2.0","ref":"readme.html#copyright-and-license","title":"Membrane RTC Engine - Copyright and License","type":"extras"},{"doc":"Apache License Version 2.0, January 2004 http://www.apache.org/licenses/ TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION 1. Definitions. &quot;License&quot; shall mean the terms and conditions for use, reproduction, and distribution as defined by Sections 1 through 9 of this document. &quot;Licensor&quot; shall mean the copyright owner or entity authorized by the copyright owner that is granting the License. &quot;Legal Entity&quot; shall mean the union of the acting entity and all other entities that control, are controlled by, or are under common control with that entity. For the purposes of this definition, &quot;control&quot; means (i) the power, direct or indirect, to cause the direction or management of such entity, whether by contract or otherwise, or (ii) ownership of fifty percent (50%) or more of the outstanding shares, or (iii) beneficial ownership of such entity. &quot;You&quot; (or &quot;Your&quot;) shall mean an individual or Legal Entity exercising permissions granted by this License. &quot;Source&quot; form shall mean the preferred form for making modifications, including but not limited to software source code, documentation source, and configuration files. &quot;Object&quot; form shall mean any form resulting from mechanical transformation or translation of a Source form, including but not limited to compiled object code, generated documentation, and conversions to other media types. &quot;Work&quot; shall mean the work of authorship, whether in Source or Object form, made available under the License, as indicated by a copyright notice that is included in or attached to the work (an example is provided in the Appendix below). &quot;Derivative Works&quot; shall mean any work, whether in Source or Object form, that is based on (or derived from) the Work and for which the editorial revisions, annotations, elaborations, or other modifications represent, as a whole, an original work of authorship. For the purposes of this License, Derivative Works shall not include works that remain separable from, or merely link (or bind by name) to the interfaces of, the Work and Derivative Works thereof. &quot;Contribution&quot; shall mean any work of authorship, including the original version of the Work and any modifications or additions to that Work or Derivative Works thereof, that is intentionally submitted to Licensor for inclusion in the Work by the copyright owner or by an individual or Legal Entity authorized to submit on behalf of the copyright owner. For the purposes of this definition, &quot;submitted&quot; means any form of electronic, verbal, or written communication sent to the Licensor or its representatives, including but not limited to communication on electronic mailing lists, source code control systems, and issue tracking systems that are managed by, or on behalf of, the Licensor for the purpose of discussing and improving the Work, but excluding communication that is conspicuously marked or otherwise designated in writing by the copyright owner as &quot;Not a Contribution.&quot; &quot;Contributor&quot; shall mean Licensor and any individual or Legal Entity on behalf of whom a Contribution has been received by Licensor and subsequently incorporated within the Work. 2. Grant of Copyright License. Subject to the terms and conditions of this License, each Contributor hereby grants to You a perpetual, worldwide, non-exclusive, no-charge, royalty-free, irrevocable copyright license to reproduce, prepare Derivative Works of, publicly display, publicly perform, sublicense, and distribute the Work and such Derivative Works in Source or Object form. 3. Grant of Patent License. Subject to the terms and conditions of this License, each Contributor hereby grants to You a perpetual, worldwide, non-exclusive, no-charge, royalty-free, irrevocable (except as stated in this section) patent license to make, have made, use, offer to sell, sell, import, and otherwise transfer the Work, where such license applies only to those patent claims licensable by such Contributor that are necessarily infringed by their Contribution(s) alone or by combination of their Contribution(s) with the Work to which such Contribution(s) was submitted. If You institute patent litigation against any entity (including a cross-claim or counterclaim in a lawsuit) alleging that the Work or a Contribution incorporated within the Work constitutes direct or contributory patent infringement, then any patent licenses granted to You under this License for that Work shall terminate as of the date such litigation is filed. 4. Redistribution. You may reproduce and distribute copies of the Work or Derivative Works thereof in any medium, with or without modifications, and in Source or Object form, provided that You meet the following conditions: (a) You must give any other recipients of the Work or Derivative Works a copy of this License; and (b) You must cause any modified files to carry prominent notices stating that You changed the files; and (c) You must retain, in the Source form of any Derivative Works that You distribute, all copyright, patent, trademark, and attribution notices from the Source form of the Work, excluding those notices that do not pertain to any part of the Derivative Works; and (d) If the Work includes a &quot;NOTICE&quot; text file as part of its distribution, then any Derivative Works that You distribute must include a readable copy of the attribution notices contained within such NOTICE file, excluding those notices that do not pertain to any part of the Derivative Works, in at least one of the following places: within a NOTICE text file distributed as part of the Derivative Works; within the Source form or documentation, if provided along with the Derivative Works; or, within a display generated by the Derivative Works, if and wherever such third-party notices normally appear. The contents of the NOTICE file are for informational purposes only and do not modify the License. You may add Your own attribution notices within Derivative Works that You distribute, alongside or as an addendum to the NOTICE text from the Work, provided that such additional attribution notices cannot be construed as modifying the License. You may add Your own copyright statement to Your modifications and may provide additional or different license terms and conditions for use, reproduction, or distribution of Your modifications, or for any such Derivative Works as a whole, provided Your use, reproduction, and distribution of the Work otherwise complies with the conditions stated in this License. 5. Submission of Contributions. Unless You explicitly state otherwise, any Contribution intentionally submitted for inclusion in the Work by You to the Licensor shall be under the terms and conditions of this License, without any additional terms or conditions. Notwithstanding the above, nothing herein shall supersede or modify the terms of any separate license agreement you may have executed with Licensor regarding such Contributions. 6. Trademarks. This License does not grant permission to use the trade names, trademarks, service marks, or product names of the Licensor, except as required for reasonable and customary use in describing the origin of the Work and reproducing the content of the NOTICE file. 7. Disclaimer of Warranty. Unless required by applicable law or agreed to in writing, Licensor provides the Work (and each Contributor provides its Contributions) on an &quot;AS IS&quot; BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied, including, without limitation, any warranties or conditions of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A PARTICULAR PURPOSE. You are solely responsible for determining the appropriateness of using or redistributing the Work and assume any risks associated with Your exercise of permissions under this License. 8. Limitation of Liability. In no event and under no legal theory, whether in tort (including negligence), contract, or otherwise, unless required by applicable law (such as deliberate and grossly negligent acts) or agreed to in writing, shall any Contributor be liable to You for damages, including any direct, indirect, special, incidental, or consequential damages of any character arising as a result of this License or out of the use or inability to use the Work (including but not limited to damages for loss of goodwill, work stoppage, computer failure or malfunction, or any and all other commercial damages or losses), even if such Contributor has been advised of the possibility of such damages. 9. Accepting Warranty or Additional Liability. While redistributing the Work or Derivative Works thereof, You may choose to offer, and charge a fee for, acceptance of support, warranty, indemnity, or other liability obligations and/or rights consistent with this License. However, in accepting such obligations, You may act only on Your own behalf and on Your sole responsibility, not on behalf of any other Contributor, and only if You agree to indemnify, defend, and hold each Contributor harmless for any liability incurred by, or claims asserted against, such Contributor by reason of your accepting any such warranty or additional liability. END OF TERMS AND CONDITIONS APPENDIX: How to apply the Apache License to your work. To apply the Apache License to your work, attach the following boilerplate notice, with the fields enclosed by brackets &quot;[]&quot; replaced with your own identifying information. (Don't include the brackets!) The text should be enclosed in the appropriate comment syntax for the file format. We also recommend that a file or class name and description of purpose be included on the same &quot;printed page&quot; as the copyright notice for easier identification within third-party archives. Copyright 2021 Software Mansion Licensed under the Apache License, Version 2.0 (the &quot;License&quot;); you may not use this file except in compliance with the License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0 Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an &quot;AS IS&quot; BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions and limitations under the License.","ref":"license.html","title":"LICENSE","type":"extras"},{"doc":"Below are described all messages, which are sent between Engine and client. Firstly there are tables with names and short descriptions. On the end, there are bullet points with a longer description and format of each media_event. Each media event have to be sent in the form: { type: name, data: data } The format described on the end represents data from the above JSON, where the name of the message is the value of the type field. RTC Engine receives these types of media_events from client: Name Description join message is sent when peer join RTC Engine leave message is sent when peer leave RTC Engine updatePeerMetadata message contains new metadata of peer updateTrackMetadata message contains new metadata of track custom message is forwarded to the endpoint associated with the peer RTC Engine sends these types of messages to the client: Name Description tracksPriority message contains all tracks that will be forwarded to peer until next tracks_priority message peerDenied message sent, if peer was rejected by server during joining to server peerUpdated message contains new metadata of peer trackUpdated message contains new metadata of track custom custom message forwarded from endpoint to client tracksAdded message sent after some tracks added by some peer tracksRemoved message contains list of tracks which are removed by some peer peerJoined message is sent after new peer joined RTC Engine peerAccepted message is sent to peer after he join RTC Engine peerLeft message is sent, when some peer left peerRemoved message is sent when peer is unwillingly removed by the server WebRTC custom media events WebRTC endpoint receives these types of custom_media_events from client: Name Description renegotiateTracks message informs that a peer wants to make renegotiation prioritizeTrack message informs about, which track peer want to prioritize unprioritizeTrack message informs about, which track peer want to unprioritize prefereVideoSizes message informs of how many videos in different quality, peer wants to receive candidate message informs about ICE candidate sdpOffer message informs all needed information for SDP negotiation WebRTC endpoint sends these type of custom messages to client Name Description offerData message informs about data needed candidate message informs about ICE candidate sdpAnswer message informs about SDP Answer join this message is sent when peer want to join RTC Engine. It contains only peer metadata { &quot;metadata&quot;: any } leave this message is sent when peer leave RTC Engine. It contains no data {} updatePeerMetadata this message contains new metadata of some peer { &quot;metadata&quot;: any } updateTrackMetadata this message contains new metadata of some track and id of this track { &quot;trackId&quot;: track_id, &quot;trackMetadata&quot;: any } custom this message is a black box for RTC Engine and it is forwarded to the endpoint associated with the peer any tracksPriority this message contains all tracks that will be forwarded to peer untill next tracks_priority message this type of messages is sent only if DisplayManager is spawned { tracks: tracks } peerDenied this message is sent, if peer was rejected by server during joining to server. It can cointains a information why he was rejected { data: text } peerUpdated this message contains information about updated metadata of one of peers { peerId: peer_id, metadata: any } trackUpdated this message contains information about new metadata of one of tracks { peerId: peer_id, trackId: track_id, metadata: any } custom this is custom message forwarded by engine from endpoint to client any tracksAdded this message informs that one of the peers add track. It contains map track_id to track_metadata of all tracks of one of the peers and id of that peer { peerId: peer_id, trackIdToMetadata: { track_id: any } } tracksRemoved this message contains list of tracks which are removed by some peer and id of that peer { peerId: peer_id, trackIds: track_ids } peerJoined this message is sent to all peers in room after new peer joined RTC Engine. It contains id and metadata of new peer. { peer: { id: peer.id, metadata: peer.metadata } } peerAccepted this message is sent to peer after he join RTC Engine. It contains his id and list of information about each peer in Engine (id, metadata and map track_id to track_metadata) { id: peer_id, peersInRoom: peers } peerLeft this message is sent to all other peers in room, when some peer left. It contains id of peer which left. { peerId: peer_id } peerRemoved this message is sent to the peer that has been unwillingly removed by the server { peerId: peer_id, reason: any } renegotiateTracks this message informs that a peer wants to make renegotiation due to adding a track or removing a track {} prioritizeTrack this message contains id of track, that peer want to prioritize { trackId: trackId } unprioritizeTrack this message contains id of track, that peer want to unprioritize { trackId: trackId } prefereVideoSizes this message contains information of how many videos in different quality (high, medium, small). There is also flag which indicates that all videos should be in the same quality. { bigScreens: Int, mediumScreens: Int, smallScreens: Int, allSameSize: boolean } candidate this message contains information about ICE candidate it is forwarded to EndpointBin { candidate: candidate, sdpMLineIndex: Int } sdpOffer this message contains SDP offer, map track_id to track_metadata and map mid to track_id. Both maps contains information only about tracks send by peer associated with this endpoint. { sdpOffer: { type: &quot;offer&quot;, sdp: sdp_offer }, trackIdToTrackMetadata: { trackId: any }, midToTrackId: mid_to_track_id } offerData this message contains information about the number of tracks of audio and video that will be sent from engine to peer and information regarding the integrated TURN server. { tracksTypes: { audio: Int, video: Int }, integratedTurnServers: [{ serverAddr: addr, serverPort: Int, transport: string, password: string, username: string }] } candidate this message contains information about ICE candidate which will be sent to the client { candidate: candidate, sdpMLineIndex: sdp_m_line_index, sdpMid: nil, usernameFragment: nil } sdpAnswer this message contains SDP answer and map mid to track_id { type: &quot;answer&quot;, sdp: sdp_answer, midToTrackId: mid_to_track_id } selectEncoding this message informs that a peer wants to receive specific encoding of some track { peerId =&gt; peer_id, trackId =&gt; track_id, encoding =&gt; encoding } encodingSwitched this message informs that track with id trackId belonging to peer with id peerId will be sent in encoding encoding now. { peerId: peer_id, trackId: track_id, encoding: encoding }","ref":"media_events.html","title":"Media Events","type":"extras"},{"doc":"Simulcast is a technique where a client sends multiple encodings of the same video to the server and the server is responsbile for choosing and forwarding proper encoding to proper receiver (other client). The encoding selection is dynamic (i.e. SFU switches between encodings in time) and it is based on: receiver awailable bandwidth receiver preferences (e.g. explicit request to receive video in HD resolution instead of FHD) UI layaout (e.g. videos being displayed in smaller video tiles will be sent in lower resolution) At the moment, Membrane supports only receiver preferences i.e. receiver can chose which encoding it is willing to receive. Additionaly, sender can turn off/on specific encoding. Membrane RTC Engine will detect changes and switch to another available encoding.","ref":"internal_simulcast.html","title":"Simulcast","type":"extras"},{"doc":"At the moment, both H264 and VP8 codecs are supported.","ref":"internal_simulcast.html#codecs","title":"Simulcast - Codecs","type":"extras"},{"doc":"Switching between encodings has to be performed transparently from the receiver point of view. This implies that: server has to rewrite some fields of RTP packets to avoid gaps in sequence numbers or timestamps after encoding switch i.e. receiver has to think it is still receiving the same RTP stream encoding switch has to be performed when server receives a keyframe from the new encoding. In other case video will freeze for a while after switching encodings.","ref":"internal_simulcast.html#logic","title":"Simulcast - Logic","type":"extras"},{"doc":"","ref":"internal_simulcast.html#architecture","title":"Simulcast - Architecture","type":"extras"},{"doc":"It is responsible for forwarding proper encoding. It encapsulates mungers (described below).","ref":"internal_simulcast.html#forwarder","title":"Simulcast - Forwarder","type":"extras"},{"doc":"Mungers are responsible for modyfing RTP packets so that switching between encodings is transparent for the receiver. Currently, there are two mungers: RTP Munger and VP8 Munger. H264 doesn't require modyfing its RTP packets so there is no H264 Munger. RTP Munger Rewrites RTP Header fields - sequence numbers and timestamps. SSRC is being rewritten by other Membrane element. VP8 Munger Rewrites VP8 RTP payload fields - keyidx, picture_id, tl0picidx","ref":"internal_simulcast.html#mungers","title":"Simulcast - Mungers","type":"extras"},{"doc":"Browser can pause sending some encoding when e.g. it doesn't have enough bandwidth. This fact is not communicated to the server. EncodingTracker is responsible for tracking encoding activity i.e. whether it is still active. In our architecture, we create EncodingTracker per simulcast encoding. Next, SimulcastTee polls information about encoding status every x seconds and when some encoding is no longer active it informs all Forwarders .","ref":"internal_simulcast.html#encodingtracker","title":"Simulcast - EncodingTracker","type":"extras"},{"doc":"Simulcast can be disabled per WebRTC Endpoint. At the moment, if client offers simulcast but we don't accept it, we will disable the whole track. This is not compliant with WebRTC standard as we should only remove SDP attributes related to simulcast and be prepared for receiving one encoding. However, in such a case browser changes SSRC after ICE restart and we cannot handle this at the moment.","ref":"internal_simulcast.html#disabling-simulcast","title":"Simulcast - Disabling Simulcast","type":"extras"},{"doc":"Simulcast is a technique where a client sends multiple encodings of the same video to the server and the server is responsbile for choosing and forwarding proper encoding to proper receiver (other client). The encoding selection is dynamic (i.e. SFU switches between encodings in time) and it is based on: receiver awailable bandwidth receiver preferences (e.g. explicit request to receive video in HD resolution instead of FHD) UI layaout (e.g. videos being displayed in smaller video tiles will be sent in a lower resolution) At the moment, Membrane supports only receiver preferences i.e. receiver can chose which encoding it is willing to receive. Additionaly, sender can turn off/on specific encoding. Membrane RTC Engine will detect changes and switch to another available encoding.","ref":"simulcast.html","title":"Simulcast","type":"extras"},{"doc":"On the client side simulcast can be enabled while adding a new track e.g.: // create MembraneWebRTC class instance // ... // add simulcasted track let trackId = webrtc.addTrack(track, stream, {}, {enabled: true, active_encodings: [&quot;l&quot;, &quot;m&quot;, &quot;h&quot;]}); This will add a new track that will be sent in three versions: original (identified as h ) original scaled down by 2 (identified as m ) original scaled down by 4 (identified as l ) You can turn off some of the encodings by excluding them from active_encodings list. Encodings that are turned off might still be enabled using enableTrackEncoding function. Minimal required resolution To make all encodings working, original track resolution has to be at least 1280x720. In other case, browser might not be able to scale resolution down. In case of browser, original track resolution can be specified in constraints passed to getUserMedia or getDisplayMedia . On the server side, simulcast can be configured while adding new WebRTC Endpoint by setting its simulcast_config option. For example % WebRTC { rtc_engine : rtc_engine , # ... simulcast_config : % SimulcastConfig { enabled : true , default_encoding : fn % Track { simulcast_encodings : _simulcast_encodings } -&gt; &quot;m&quot; end } } Here we turn simulcast on and choose medium encoding for each track to be forwarded to the client. On the other hand, setting enabled to false will result in rejecting all incoming simulcast tracks i.e. client will not send them to the server.","ref":"simulcast.html#turning-simulcast-on-off","title":"Simulcast - Turning simulcast on/off","type":"extras"},{"doc":"After adding simulcasted track, user can disable specific encoding. For example, to disable the highest encoding: // create MembraneWebRTC class instance // ... // add simulcasted track let trackId = webrtc.addTrack(track, stream, {}, true); webrtc.disableTrackEncoding(trackId, &quot;h&quot;); Disabled encoding can be turned on again using enableTrackEncoding function. Membrane RTC Engine tracks encoding activity. Therefore, when some encoding is turned off, RTC Engine will detect this and switch to the highest awailable encoding. When disabled encoding becomes active again, RTC Engine will switch back to it.","ref":"simulcast.html#disabling-and-enabling-specific-track-encoding","title":"Simulcast - Disabling and enabling specific track encoding","type":"extras"},{"doc":"By default, RTC Engine will forward the highest available encoding. However, this can be changed using selectTrackEncoding function. For example, to receive the lowest resolution: webrtc.selectTrackEncoding(peerId, trackId, &quot;l&quot;); where peerId is id of peer this track belongs to.","ref":"simulcast.html#selecting-encoding-to-receive","title":"Simulcast - Selecting encoding to receive","type":"extras"}]